package cyclops.collections.mutable;

import com.aol.cyclops2.data.collections.extensions.lazy.LazySortedSetX;
import com.aol.cyclops2.types.Zippable;
import com.aol.cyclops2.types.anyM.AnyMSeq;
import com.aol.cyclops2.types.foldable.ConvertableSequence;
import com.aol.cyclops2.types.foldable.Evaluation;

import com.aol.cyclops2.data.collections.extensions.standard.LazyCollectionX;
import cyclops.collections.immutable.VectorX;
import cyclops.control.Xor;
import cyclops.function.Monoid;
import cyclops.monads.AnyM;
import cyclops.monads.Witness.sortedSet;
import cyclops.stream.ReactiveSeq;
import cyclops.control.Trampoline;
import com.aol.cyclops2.types.recoverable.OnEmptySwitch;
import com.aol.cyclops2.types.foldable.To;
import cyclops.function.Fn3;
import cyclops.function.Fn4;
import cyclops.stream.Spouts;
import org.jooq.lambda.Collectable;
import org.jooq.lambda.Seq;
import org.jooq.lambda.tuple.Tuple2;
import org.jooq.lambda.tuple.Tuple3;
import org.jooq.lambda.tuple.Tuple4;
import org.reactivestreams.Publisher;

import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.util.*;
import java.util.concurrent.TimeUnit;
import java.util.function.*;
import java.util.stream.Collector;
import java.util.stream.Collectors;
import java.util.stream.Stream;
import static com.aol.cyclops2.types.foldable.Evaluation.LAZY;

/**
 * An eXtended SortedSet type, that offers additional functional style operators such as bimap, filter and more
 * Can operate eagerly, lazily or reactively (async push)
 *
 * @author johnmcclean
 *
 * @param <T> the type of elements held in this SortedSetX
 */
public interface SortedSetX<T> extends To<SortedSetX<T>>,SortedSet<T>, LazyCollectionX<T>, OnEmptySwitch<T, SortedSet<T>> {


    public static <T> SortedSetX<T> defer(Supplier<SortedSetX<T>> s){
        return of(s)
                .map(Supplier::get)
                .flatMap(l->l);
    }

    static <T> CompletableSortedSetX<T> completable(){
        return new CompletableSortedSetX<>();
    }

    static class CompletableSortedSetX<T> implements InvocationHandler {
        cyclops.async.Future<SortedSetX<T>> future = cyclops.async.Future.future();
        public boolean complete(SortedSet<T> result){
            return future.complete(SortedSetX.fromIterable(result));
        }

        public SortedSetX<T> asSortedSetX(){
            SortedSetX f = (SortedSetX) Proxy.newProxyInstance(SortedSetX.class.getClassLoader(),
                    new Class[] { SortedSetX.class },
                    this);
            return f;
        }

        @Override
        public Object invoke(Object proxy, Method method, Object[] args) throws Throwable {
            SortedSetX<T> target = future.get();
            return method.invoke(target,args);
        }
    }
    SortedSetX<T> lazy();
    SortedSetX<T> eager();
    static <T> Collector<T, ?, SortedSet<T>> defaultCollector() {
        return Collectors.toCollection(() -> new TreeSet<T>(
                                                            (Comparator) Comparator.<Comparable> naturalOrder()));
    }

    static <T> Collector<T, ?, SortedSet<T>> immutableCollector() {
        return Collectors.collectingAndThen(defaultCollector(), (final SortedSet<T> d) -> Collections.unmodifiableSortedSet(d));

    }

    /**
    * Create a SortedSetX that contains the Integers between skip and take
    * 
    * @param start
    *            Number of range to skip from
    * @param end
    *            Number for range to take at
    * @return Range SortedSetX
    */
    public static SortedSetX<Integer> range(final int start, final int end) {
        return ReactiveSeq.range(start, end)
                          .to()
                          .sortedSetX(Evaluation.LAZY);
    }

    /**
     * Create a SortedSetX that contains the Longs between skip and take
     * 
     * @param start
     *            Number of range to skip from
     * @param end
     *            Number for range to take at
     * @return Range SortedSetX
     */
    public static SortedSetX<Long> rangeLong(final long start, final long end) {
        return ReactiveSeq.rangeLong(start, end)
                          .to()
                          .sortedSetX(Evaluation.LAZY);
    }

    /**
     * Unfold a function into a SortedSetX
     * 
     * <pre>
     * {@code 
     *  SortedSetX.unfold(1,i->i<=6 ? Optional.of(Tuple.tuple(i,i+1)) : Optional.zero());
     * 
     * //(1,2,3,4,5)
     * 
     * }</pre>
     * 
     * @param seed Initial value 
     * @param unfolder Iteratively applied function, terminated by an zero Optional
     * @return SortedSetX generated by unfolder function
     */
    static <U, T> SortedSetX<T> unfold(final U seed, final Function<? super U, Optional<Tuple2<T, U>>> unfolder) {
        return ReactiveSeq.unfold(seed, unfolder)
                          .to()
                          .sortedSetX(Evaluation.LAZY);
    }

    /**
     * Generate a SortedSetX from the provided Supplier up to the provided limit number of times
     * 
     * @param limit Max number of elements to generate
     * @param s Supplier to generate SortedSetX elements
     * @return SortedSetX generated from the provided Supplier
     */
    public static <T> SortedSetX<T> generate(final long limit, final Supplier<T> s) {

        return ReactiveSeq.generate(s)
                          .limit(limit)
                          .to()
                          .sortedSetX(Evaluation.LAZY);
    }

    /**
     * Create a SortedSetX by iterative application of a function to an initial element up to the supplied limit number of times
     * 
     * @param limit Max number of elements to generate
     * @param seed Initial element
     * @param f Iteratively applied to each element to generate the next element
     * @return SortedSetX generated by iterative application
     */
    public static <T> SortedSetX<T> iterate(final long limit, final T seed, final UnaryOperator<T> f) {
        return ReactiveSeq.iterate(seed, f)
                          .limit(limit)
                          .to()
                          .sortedSetX(Evaluation.LAZY);

    }

    public static <T> SortedSetX<T> empty() {
        return fromIterable((SortedSet<T>) defaultCollector().supplier()
                                                             .get());
    }

    public static <T> SortedSetX<T> of(final T... values) {
        return new LazySortedSetX<T>(null,
                ReactiveSeq.of(values),
                defaultCollector(),Evaluation.LAZY);
    }

    public static <T> SortedSetX<T> singleton(final T value) {
        return of(value);
    }

    @Override
    default SortedSetX<T> materialize() {
        return (SortedSetX<T>)LazyCollectionX.super.materialize();
    }


    @Override
    default SortedSetX<T> take(final long num) {

        return (SortedSetX<T>) LazyCollectionX.super.limit(num);
    }
    @Override
    default SortedSetX<T> drop(final long num) {

        return (SortedSetX<T>) LazyCollectionX.super.skip(num);
    }
    /**
     * Construct a SortedSetX from an Publisher
     * 
     * @param publisher
     *            to construct SortedSetX from
     * @return SortedSetX
     */
    public static <T> SortedSetX<T> fromPublisher(final Publisher<? extends T> publisher) {
        return Spouts.from((Publisher<T>) publisher)
                          .to()
                          .sortedSetX(Evaluation.LAZY);
    }
    /**
     *
     * <pre>
     * {@code
     *  import static cyclops.stream.ReactiveSeq.range;
     *
     *  SortedSetX<Integer> sortedSet = sortedSetX(range(10,20));
     *
     * }
     * </pre>
     * @param stream To create SortedSetX from
     * @param <T> SortedSetX generated from Stream
     * @return
     */
    public static <T> SortedSetX<T> sortedSetX(ReactiveSeq<T> stream){
        return new LazySortedSetX<T>(null,
                stream,
                defaultCollector(),Evaluation.LAZY);
    }

    public static <T> SortedSetX<T> fromIterable(final Iterable<T> it) {
        if (it instanceof SortedSetX)
            return (SortedSetX<T>) it;
        if (it instanceof SortedSet)
            return new LazySortedSetX<T>(
                                         (SortedSet) it, defaultCollector(),Evaluation.LAZY);
        return new LazySortedSetX<T>(null,
                                       ReactiveSeq.fromIterable(it),
                                     defaultCollector(),Evaluation.LAZY);
    }
    public static <T> SortedSetX<T> fromIterator(final Iterator<T> it) {
        return fromIterable(()->it);
    }

    public static <T> SortedSetX<T> fromIterable(final Collector<T, ?, SortedSet<T>> collector, final Iterable<T> it) {
        if (it instanceof SortedSetX)
            return ((SortedSetX<T>) it).type(collector);
        if (it instanceof SortedSet)
            return new LazySortedSetX<T>(
                                         (SortedSet) it, collector,Evaluation.LAZY);
        return new LazySortedSetX<T>(null,
                                     ReactiveSeq.fromIterable(it),
                                     collector,Evaluation.LAZY);
    }
    
    SortedSetX<T> type(Collector<T, ?, SortedSet<T>> collector);
    default AnyMSeq<sortedSet,T> anyM(){
        return AnyM.fromSortedSet(this);
    }
    /**
     * coflatMap pattern, can be used to perform lazy reductions / collections / folds and other terminal operations
     * 
     * <pre>
     * {@code 
     *   
     *     SortedSetX.of(1,2,3)
     *               .map(i->i*2)
     *               .coflatMap(s -> s.reduce(0,(a,b)->a+b))
     *      
     *      //SortedSetX[12]
     * }
     * </pre>
     * 
     * 
     * @param fn mapping function
     * @return Transformed SortedSet
     */
    default <R> SortedSetX<R> coflatMap(Function<? super SortedSetX<T>, ? extends R> fn){
        return fn.andThen(r ->  this.<R>unit(r))
                .apply(this);
    }


    /* (non-Javadoc)
     * @see com.aol.cyclops2.data.collections.extensions.CollectionX#forEach4(java.util.function.Function, java.util.function.BiFunction, com.aol.cyclops2.util.function.TriFunction, com.aol.cyclops2.util.function.QuadFunction)
     */
    @Override
    default <R1, R2, R3, R> SortedSetX<R> forEach4(Function<? super T, ? extends Iterable<R1>> stream1,
            BiFunction<? super T, ? super R1, ? extends Iterable<R2>> stream2,
            Fn3<? super T, ? super R1, ? super R2, ? extends Iterable<R3>> stream3,
            Fn4<? super T, ? super R1, ? super R2, ? super R3, ? extends R> yieldingFunction) {
        
        return (SortedSetX)LazyCollectionX.super.forEach4(stream1, stream2, stream3, yieldingFunction);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.data.collections.extensions.CollectionX#forEach4(java.util.function.Function, java.util.function.BiFunction, com.aol.cyclops2.util.function.TriFunction, com.aol.cyclops2.util.function.QuadFunction, com.aol.cyclops2.util.function.QuadFunction)
     */
    @Override
    default <R1, R2, R3, R> SortedSetX<R> forEach4(Function<? super T, ? extends Iterable<R1>> stream1,
            BiFunction<? super T, ? super R1, ? extends Iterable<R2>> stream2,
            Fn3<? super T, ? super R1, ? super R2, ? extends Iterable<R3>> stream3,
            Fn4<? super T, ? super R1, ? super R2, ? super R3, Boolean> filterFunction,
            Fn4<? super T, ? super R1, ? super R2, ? super R3, ? extends R> yieldingFunction) {
        
        return (SortedSetX)LazyCollectionX.super.forEach4(stream1, stream2, stream3, filterFunction, yieldingFunction);
    }
    /**
    default ConvertableSequence<T> to(){


        return new ConvertableSequence<>(this);
    }
    default Collectable<T> collectors(){



        return Seq.seq(this);
    }**/
    /* (non-Javadoc)
     * @see com.aol.cyclops2.data.collections.extensions.CollectionX#forEach3(java.util.function.Function, java.util.function.BiFunction, com.aol.cyclops2.util.function.TriFunction)
     */
    @Override
    default <R1, R2, R> SortedSetX<R> forEach3(Function<? super T, ? extends Iterable<R1>> stream1,
            BiFunction<? super T, ? super R1, ? extends Iterable<R2>> stream2,
            Fn3<? super T, ? super R1, ? super R2, ? extends R> yieldingFunction) {
        
        return (SortedSetX)LazyCollectionX.super.forEach3(stream1, stream2, yieldingFunction);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.data.collections.extensions.CollectionX#forEach3(java.util.function.Function, java.util.function.BiFunction, com.aol.cyclops2.util.function.TriFunction, com.aol.cyclops2.util.function.TriFunction)
     */
    @Override
    default <R1, R2, R> SortedSetX<R> forEach3(Function<? super T, ? extends Iterable<R1>> stream1,
            BiFunction<? super T, ? super R1, ? extends Iterable<R2>> stream2,
            Fn3<? super T, ? super R1, ? super R2, Boolean> filterFunction,
            Fn3<? super T, ? super R1, ? super R2, ? extends R> yieldingFunction) {
        
        return (SortedSetX)LazyCollectionX.super.forEach3(stream1, stream2, filterFunction, yieldingFunction);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.data.collections.extensions.CollectionX#forEach2(java.util.function.Function, java.util.function.BiFunction)
     */
    @Override
    default <R1, R> SortedSetX<R> forEach2(Function<? super T, ? extends Iterable<R1>> stream1,
            BiFunction<? super T, ? super R1, ? extends R> yieldingFunction) {
        
        return (SortedSetX)LazyCollectionX.super.forEach2(stream1, yieldingFunction);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.data.collections.extensions.CollectionX#forEach2(java.util.function.Function, java.util.function.BiFunction, java.util.function.BiFunction)
     */
    @Override
    default <R1, R> SortedSetX<R> forEach2(Function<? super T, ? extends Iterable<R1>> stream1,
            BiFunction<? super T, ? super R1, Boolean> filterFunction,
            BiFunction<? super T, ? super R1, ? extends R> yieldingFunction) {
        
        return (SortedSetX)LazyCollectionX.super.forEach2(stream1, filterFunction, yieldingFunction);
    }
    

    /**
     * Combine two adjacent elements in a SortedSetX using the supplied BinaryOperator
     * This is a stateful grouping and reduction operation. The emitted of a combination may in turn be combined
     * with it's neighbor
     * <pre>
     * {@code 
     *  SortedSetX.of(1,1,2,3)
                   .combine((a, b)->a.equals(b),SemigroupK.intSum)
                   .listX(Evaluation.LAZY)
                   
     *  //ListX(3,4) 
     * }</pre>
     * 
     * @param predicate Test to see if two neighbors should be joined
     * @param op Reducer to combine neighbors
     * @return Combined / Partially Reduced SortedSetX
     */
    @Override
    default SortedSetX<T> combine(final BiPredicate<? super T, ? super T> predicate, final BinaryOperator<T> op) {
        return (SortedSetX<T>) LazyCollectionX.super.combine(predicate, op);
    }
    @Override
    default SortedSetX<T> combine(final Monoid<T> op, final BiPredicate<? super T, ? super T> predicate) {
        return (SortedSetX<T>)LazyCollectionX.super.combine(op,predicate);
    }
    @Override
    default <R> SortedSetX<R> unit(final R value) {
        return singleton(value);
    }

    @Override
    default <R> SortedSetX<R> unit(final Collection<R> col) {
        return fromIterable(col);
    }

    @Override
    default <R> SortedSetX<R> unitIterator(final Iterator<R> it) {
        return fromIterable(() -> it);
    }

    @Override
    default ReactiveSeq<T> stream() {

        return ReactiveSeq.fromIterable(this);
    }

    @Override
    default <T1> SortedSetX<T1> from(final Collection<T1> c) {
        return SortedSetX.<T1> fromIterable(getCollector(), c);
    }

    public <T> Collector<T, ?, SortedSet<T>> getCollector();

    @Override
    default <X> SortedSetX<X> fromStream(final ReactiveSeq<X> stream) {
        return new LazySortedSetX<>(
                                    stream.collect(getCollector()), getCollector(),Evaluation.LAZY);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.standard.LazyCollectionX#reverse()
     */
    @Override
    default SortedSetX<T> reverse() {
        return (SortedSetX<T>) LazyCollectionX.super.reverse();
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.standard.LazyCollectionX#filter(java.util.function.Predicate)
     */
    @Override
    default SortedSetX<T> filter(final Predicate<? super T> pred) {

        return (SortedSetX<T>) LazyCollectionX.super.filter(pred);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.standard.LazyCollectionX#map(java.util.function.Function)
     */
    @Override
    default <R> SortedSetX<R> map(final Function<? super T, ? extends R> mapper) {

        return (SortedSetX<R>) LazyCollectionX.super.<R> map(mapper);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.standard.LazyCollectionX#flatMap(java.util.function.Function)
     */
    @Override
    default <R> SortedSetX<R> flatMap(final Function<? super T, ? extends Iterable<? extends R>> mapper) {

        return (SortedSetX<R>) LazyCollectionX.super.<R> flatMap(mapper);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.standard.LazyCollectionX#limit(long)
     */
    @Override
    default SortedSetX<T> limit(final long num) {
        return (SortedSetX<T>) LazyCollectionX.super.limit(num);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.standard.LazyCollectionX#skip(long)
     */
    @Override
    default SortedSetX<T> skip(final long num) {

        return (SortedSetX<T>) LazyCollectionX.super.skip(num);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.standard.LazyCollectionX#takeWhile(java.util.function.Predicate)
     */
    @Override
    default SortedSetX<T> takeWhile(final Predicate<? super T> p) {

        return (SortedSetX<T>) LazyCollectionX.super.takeWhile(p);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.standard.LazyCollectionX#dropWhile(java.util.function.Predicate)
     */
    @Override
    default SortedSetX<T> dropWhile(final Predicate<? super T> p) {

        return (SortedSetX<T>) LazyCollectionX.super.dropWhile(p);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.standard.LazyCollectionX#takeUntil(java.util.function.Predicate)
     */
    @Override
    default SortedSetX<T> takeUntil(final Predicate<? super T> p) {

        return (SortedSetX<T>) LazyCollectionX.super.takeUntil(p);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.standard.LazyCollectionX#dropUntil(java.util.function.Predicate)
     */
    @Override
    default SortedSetX<T> dropUntil(final Predicate<? super T> p) {

        return (SortedSetX<T>) LazyCollectionX.super.dropUntil(p);
    }

    @Override
    default SortedSetX<T> takeRight(final int num) {
        return (SortedSetX<T>) LazyCollectionX.super.takeRight(num);
    }

    @Override
    default SortedSetX<T> dropRight(final int num) {
        return (SortedSetX<T>) LazyCollectionX.super.dropRight(num);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.standard.LazyCollectionX#trampoline(java.util.function.Function)
     */
    @Override
    default <R> SortedSetX<R> trampoline(final Function<? super T, ? extends Trampoline<? extends R>> mapper) {

        return (SortedSetX<R>) LazyCollectionX.super.<R> trampoline(mapper);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.standard.LazyCollectionX#slice(long, long)
     */
    @Override
    default SortedSetX<T> slice(final long from, final long to) {

        return (SortedSetX<T>) LazyCollectionX.super.slice(from, to);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.standard.LazyCollectionX#sorted(java.util.function.Function)
     */
    @Override
    default <U extends Comparable<? super U>> SortedSetX<T> sorted(final Function<? super T, ? extends U> function) {

        return (SortedSetX<T>) LazyCollectionX.super.sorted(function);
    }

    @Override
    default SortedSetX<ListX<T>> grouped(final int groupSize) {
        return (SortedSetX<ListX<T>>) (SortedSetX<T>) LazyCollectionX.super.grouped(groupSize);
    }

    @Override
    default <K, A, D> SortedSetX<Tuple2<K, D>> grouped(final Function<? super T, ? extends K> classifier,
            final Collector<? super T, A, D> downstream) {
        return (SortedSetX) LazyCollectionX.super.grouped(classifier, downstream);
    }

    @Override
    default <K> SortedSetX<Tuple2<K, ReactiveSeq<T>>> grouped(final Function<? super T, ? extends K> classifier) {

        return (SortedSetX) fromStream(stream().grouped(classifier)
                                               .map(t -> t.map2(Comparables::comparable)));
    }

    @Override
    default <U> SortedSetX<Tuple2<T, U>> zip(final Iterable<? extends U> other) {
        return (SortedSetX<Tuple2<T, U>>) (SortedSetX<T>) LazyCollectionX.super.zip(other);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.standard.LazyCollectionX#zip(java.lang.Iterable, java.util.function.BiFunction)
     */
    @Override
    default <U, R> SortedSetX<R> zip(final Iterable<? extends U> other, final BiFunction<? super T, ? super U, ? extends R> zipper) {
        return (SortedSetX<R>) LazyCollectionX.super.zip(other, zipper);
    }


    @Override
    default <U, R> SortedSetX<R> zipS(final Stream<? extends U> other, final BiFunction<? super T, ? super U, ? extends R> zipper) {
        return (SortedSetX<R>) LazyCollectionX.super.zipS(other, zipper);
    }

    @Override
    default SortedSetX<VectorX<T>> sliding(final int windowSize) {
        return (SortedSetX<VectorX<T>>) (SortedSetX<T>) LazyCollectionX.super.sliding(windowSize);
    }

    @Override
    default SortedSetX<VectorX<T>> sliding(final int windowSize, final int increment) {
        return (SortedSetX<VectorX<T>>) (SortedSetX<T>) LazyCollectionX.super.sliding(windowSize, increment);
    }

    @Override
    default SortedSetX<T> scanLeft(final Monoid<T> monoid) {
        return (SortedSetX<T>) LazyCollectionX.super.scanLeft(monoid);
    }

    @Override
    default <U> SortedSetX<U> scanLeft(final U seed, final BiFunction<? super U, ? super T, ? extends U> function) {
        return (SortedSetX<U>) (SortedSetX<T>) LazyCollectionX.super.scanLeft(seed, function);
    }

    @Override
    default SortedSetX<T> scanRight(final Monoid<T> monoid) {
        return (SortedSetX<T>) LazyCollectionX.super.scanRight(monoid);
    }

    @Override
    default <U> SortedSetX<U> scanRight(final U identity, final BiFunction<? super T, ? super U, ? extends U> combiner) {
        return (SortedSetX<U>) (SortedSetX<T>) LazyCollectionX.super.scanRight(identity, combiner);
    }

    @Override
    default SortedSetX<T> plus(final T e) {
        add(e);
        return this;
    }

    @Override
    default SortedSetX<T> plusAll(final Collection<? extends T> list) {
        addAll(list);
        return this;
    }

    @Override
    default SortedSetX<T> minus(final Object e) {
        remove(e);
        return this;
    }

    @Override
    default SortedSetX<T> minusAll(final Collection<?> list) {
        removeAll(list);
        return this;
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.standard.LazyCollectionX#cycle(int)
     */
    @Override
    default ListX<T> cycle(final long times) {

        return this.stream()
                   .cycle(times)
                   .to().listX(LAZY);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.standard.LazyCollectionX#cycle(com.aol.cyclops2.sequence.Monoid, int)
     */
    @Override
    default ListX<T> cycle(final Monoid<T> m, final long times) {

        return this.stream()
                   .cycle(m, times)
                   .to().listX(LAZY);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.standard.LazyCollectionX#cycleWhile(java.util.function.Predicate)
     */
    @Override
    default ListX<T> cycleWhile(final Predicate<? super T> predicate) {

        return this.stream()
                   .cycleWhile(predicate)
                   .to().listX(LAZY);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.standard.LazyCollectionX#cycleUntil(java.util.function.Predicate)
     */
    @Override
    default ListX<T> cycleUntil(final Predicate<? super T> predicate) {

        return this.stream()
                   .cycleUntil(predicate)
                   .to().listX(LAZY);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.standard.LazyCollectionX#zip(java.util.reactiveStream.Stream)
     */
    @Override
    default <U> SortedSetX<Tuple2<T, U>> zipS(final Stream<? extends U> other) {

        return (SortedSetX) LazyCollectionX.super.zipS(other);
    }



    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.standard.LazyCollectionX#zip3(java.util.reactiveStream.Stream, java.util.reactiveStream.Stream)
     */
    @Override
    default <S, U> SortedSetX<Tuple3<T, S, U>> zip3(final Iterable<? extends S> second, final Iterable<? extends U> third) {

        return (SortedSetX) LazyCollectionX.super.zip3(second, third);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.standard.LazyCollectionX#zip4(java.util.reactiveStream.Stream, java.util.reactiveStream.Stream, java.util.reactiveStream.Stream)
     */
    @Override
    default <T2, T3, T4> SortedSetX<Tuple4<T, T2, T3, T4>> zip4(final Iterable<? extends T2> second, final Iterable<? extends T3> third,
            final Iterable<? extends T4> fourth) {

        return (SortedSetX) LazyCollectionX.super.zip4(second, third, fourth);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.standard.LazyCollectionX#zipWithIndex()
     */
    @Override
    default SortedSetX<Tuple2<T, Long>> zipWithIndex() {

        return (SortedSetX<Tuple2<T, Long>>) LazyCollectionX.super.zipWithIndex();
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.standard.LazyCollectionX#distinct()
     */
    @Override
    default SortedSetX<T> distinct() {

        return (SortedSetX<T>) LazyCollectionX.super.distinct();
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.standard.LazyCollectionX#sorted()
     */
    @Override
    default SortedSetX<T> sorted() {

        return (SortedSetX<T>) LazyCollectionX.super.sorted();
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.standard.LazyCollectionX#sorted(java.util.Comparator)
     */
    @Override
    default SortedSetX<T> sorted(final Comparator<? super T> c) {

        return (SortedSetX<T>) LazyCollectionX.super.sorted(c);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.standard.LazyCollectionX#skipWhile(java.util.function.Predicate)
     */
    @Override
    default SortedSetX<T> skipWhile(final Predicate<? super T> p) {

        return (SortedSetX<T>) LazyCollectionX.super.skipWhile(p);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.standard.LazyCollectionX#skipUntil(java.util.function.Predicate)
     */
    @Override
    default SortedSetX<T> skipUntil(final Predicate<? super T> p) {

        return (SortedSetX<T>) LazyCollectionX.super.skipUntil(p);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.standard.LazyCollectionX#limitWhile(java.util.function.Predicate)
     */
    @Override
    default SortedSetX<T> limitWhile(final Predicate<? super T> p) {

        return (SortedSetX<T>) LazyCollectionX.super.limitWhile(p);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.standard.LazyCollectionX#limitUntil(java.util.function.Predicate)
     */
    @Override
    default SortedSetX<T> limitUntil(final Predicate<? super T> p) {

        return (SortedSetX<T>) LazyCollectionX.super.limitUntil(p);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.standard.LazyCollectionX#intersperse(java.lang.Object)
     */
    @Override
    default SortedSetX<T> intersperse(final T value) {

        return (SortedSetX<T>) LazyCollectionX.super.intersperse(value);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.standard.LazyCollectionX#shuffle()
     */
    @Override
    default SortedSetX<T> shuffle() {

        return (SortedSetX<T>) LazyCollectionX.super.shuffle();
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.standard.LazyCollectionX#skipLast(int)
     */
    @Override
    default SortedSetX<T> skipLast(final int num) {

        return (SortedSetX<T>) LazyCollectionX.super.skipLast(num);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.standard.LazyCollectionX#limitLast(int)
     */
    @Override
    default SortedSetX<T> limitLast(final int num) {

        return (SortedSetX<T>) LazyCollectionX.super.limitLast(num);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.types.recoverable.OnEmptySwitch#onEmptySwitch(java.util.function.Supplier)
     */
    @Override
    default SortedSetX<T> onEmptySwitch(final Supplier<? extends SortedSet<T>> supplier) {
        if (isEmpty())
            return SortedSetX.fromIterable(supplier.get());
        return this;
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.standard.LazyCollectionX#onEmpty(java.lang.Object)
     */
    @Override
    default SortedSetX<T> onEmpty(final T value) {

        return (SortedSetX<T>) LazyCollectionX.super.onEmpty(value);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.standard.LazyCollectionX#onEmptyGet(java.util.function.Supplier)
     */
    @Override
    default SortedSetX<T> onEmptyGet(final Supplier<? extends T> supplier) {

        return (SortedSetX<T>) LazyCollectionX.super.onEmptyGet(supplier);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.standard.LazyCollectionX#onEmptyThrow(java.util.function.Supplier)
     */
    @Override
    default <X extends Throwable> SortedSetX<T> onEmptyThrow(final Supplier<? extends X> supplier) {

        return (SortedSetX<T>) LazyCollectionX.super.onEmptyThrow(supplier);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.standard.LazyCollectionX#shuffle(java.util.Random)
     */
    @Override
    default SortedSetX<T> shuffle(final Random random) {

        return (SortedSetX<T>) LazyCollectionX.super.shuffle(random);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.standard.LazyCollectionX#ofType(java.lang.Class)
     */
    @Override
    default <U> SortedSetX<U> ofType(final Class<? extends U> type) {

        return (SortedSetX<U>) LazyCollectionX.super.ofType(type);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.standard.LazyCollectionX#filterNot(java.util.function.Predicate)
     */
    @Override
    default SortedSetX<T> filterNot(final Predicate<? super T> fn) {

        return (SortedSetX<T>) LazyCollectionX.super.filterNot(fn);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.standard.LazyCollectionX#notNull()
     */
    @Override
    default SortedSetX<T> notNull() {

        return (SortedSetX<T>) LazyCollectionX.super.notNull();
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.standard.LazyCollectionX#removeAll(java.util.reactiveStream.Stream)
     */
    @Override
    default SortedSetX<T> removeAllS(final Stream<? extends T> stream) {

        return (SortedSetX<T>) LazyCollectionX.super.removeAllS(stream);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.standard.LazyCollectionX#removeAll(java.lang.Iterable)
     */
    @Override
    default SortedSetX<T> removeAllI(final Iterable<? extends T> it) {

        return (SortedSetX<T>) LazyCollectionX.super.removeAllI(it);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.standard.LazyCollectionX#removeAll(java.lang.Object[])
     */
    @Override
    default SortedSetX<T> removeAll(final T... values) {

        return (SortedSetX<T>) LazyCollectionX.super.removeAll(values);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.standard.LazyCollectionX#retainAllI(java.lang.Iterable)
     */
    @Override
    default SortedSetX<T> retainAllI(final Iterable<? extends T> it) {

        return (SortedSetX<T>) LazyCollectionX.super.retainAllI(it);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.standard.LazyCollectionX#retainAllI(java.util.reactiveStream.Stream)
     */
    @Override
    default SortedSetX<T> retainAllS(final Stream<? extends T> seq) {

        return (SortedSetX<T>) LazyCollectionX.super.retainAllS(seq);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.standard.LazyCollectionX#retainAllI(java.lang.Object[])
     */
    @Override
    default SortedSetX<T> retainAll(final T... values) {

        return (SortedSetX<T>) LazyCollectionX.super.retainAll(values);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.standard.LazyCollectionX#cast(java.lang.Class)
     */
    @Override
    default <U> SortedSetX<U> cast(final Class<? extends U> type) {

        return (SortedSetX<U>) LazyCollectionX.super.cast(type);
    }


    @Override
    default <C extends Collection<? super T>> SortedSetX<C> grouped(final int size, final Supplier<C> supplier) {

        return (SortedSetX<C>) LazyCollectionX.super.grouped(size, supplier);
    }

    @Override
    default SortedSetX<ListX<T>> groupedUntil(final Predicate<? super T> predicate) {

        return (SortedSetX<ListX<T>>) LazyCollectionX.super.groupedUntil(predicate);
    }

    @Override
    default SortedSetX<ListX<T>> groupedWhile(final Predicate<? super T> predicate) {

        return (SortedSetX<ListX<T>>) LazyCollectionX.super.groupedWhile(predicate);
    }

    @Override
    default <C extends Collection<? super T>> SortedSetX<C> groupedWhile(final Predicate<? super T> predicate, final Supplier<C> factory) {

        return (SortedSetX<C>) LazyCollectionX.super.groupedWhile(predicate, factory);
    }

    @Override
    default <C extends Collection<? super T>> SortedSetX<C> groupedUntil(final Predicate<? super T> predicate, final Supplier<C> factory) {

        return (SortedSetX<C>) LazyCollectionX.super.groupedUntil(predicate, factory);
    }

    @Override
    default SortedSetX<ListX<T>> groupedStatefullyUntil(final BiPredicate<ListX<? super T>, ? super T> predicate) {

        return (SortedSetX<ListX<T>>) LazyCollectionX.super.groupedStatefullyUntil(predicate);
    }


    /* (non-Javadoc)
     * @see com.aol.cyclops2.lambda.monads.ExtendedTraversable#permutations()
     */
    @Override
    default SortedSetX<ReactiveSeq<T>> permutations() {
        return fromStream(stream().permutations()
                                  .map(Comparables::comparable));

    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.lambda.monads.ExtendedTraversable#combinations(int)
     */
    @Override
    default SortedSetX<ReactiveSeq<T>> combinations(final int size) {
        return fromStream(stream().combinations(size)
                                  .map(Comparables::comparable));
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.lambda.monads.ExtendedTraversable#combinations()
     */
    @Override
    default SortedSetX<ReactiveSeq<T>> combinations() {
        return fromStream(stream().combinations()
                                  .map(Comparables::comparable));
    }


    @Override
    default <R> SortedSetX<R> retry(final Function<? super T, ? extends R> fn) {
        return (SortedSetX<R>)LazyCollectionX.super.retry(fn);
    }

    @Override
    default <R> SortedSetX<R> retry(final Function<? super T, ? extends R> fn, final int retries, final long delay, final TimeUnit timeUnit) {
        return (SortedSetX<R>)LazyCollectionX.super.retry(fn);
    }

    @Override
    default <R> SortedSetX<R> flatMapS(Function<? super T, ? extends Stream<? extends R>> fn) {
        return (SortedSetX<R>)LazyCollectionX.super.flatMapS(fn);
    }

    @Override
    default <R> SortedSetX<R> flatMapP(Function<? super T, ? extends Publisher<? extends R>> fn) {
        return (SortedSetX<R>)LazyCollectionX.super.flatMapP(fn);
    }

    @Override
    default SortedSetX<T> prependS(Stream<? extends T> stream) {
        return (SortedSetX<T>)LazyCollectionX.super.prependS(stream);
    }

    @Override
    default SortedSetX<T> append(T... values) {
        return (SortedSetX<T>)LazyCollectionX.super.append(values);
    }

    @Override
    default SortedSetX<T> append(T value) {
        return (SortedSetX<T>)LazyCollectionX.super.append(value);
    }

    @Override
    default SortedSetX<T> prepend(T value) {
        return (SortedSetX<T>)LazyCollectionX.super.prepend(value);
    }

    @Override
    default SortedSetX<T> prepend(T... values) {
        return (SortedSetX<T>)LazyCollectionX.super.prepend(values);
    }

    @Override
    default SortedSetX<T> insertAt(int pos, T... values) {
        return (SortedSetX<T>)LazyCollectionX.super.insertAt(pos,values);
    }

    @Override
    default SortedSetX<T> deleteBetween(int start, int end) {
        return (SortedSetX<T>)LazyCollectionX.super.deleteBetween(start,end);
    }

    @Override
    default SortedSetX<T> insertAtS(int pos, Stream<T> stream) {
        return (SortedSetX<T>)LazyCollectionX.super.insertAtS(pos,stream);
    }

    @Override
    default SortedSetX<T> recover(final Function<? super Throwable, ? extends T> fn) {
        return (SortedSetX<T>)LazyCollectionX.super.recover(fn);
    }

    @Override
    default <EX extends Throwable> SortedSetX<T> recover(Class<EX> exceptionClass, final Function<? super EX, ? extends T> fn) {
        return (SortedSetX<T>)LazyCollectionX.super.recover(exceptionClass,fn);
    }
    @Override
    default SortedSetX<T> plusLoop(int max, IntFunction<T> value) {
        return (SortedSetX<T>)LazyCollectionX.super.plusLoop(max,value);
    }

    @Override
    default SortedSetX<T> plusLoop(Supplier<Optional<T>> supplier) {
        return (SortedSetX<T>)LazyCollectionX.super.plusLoop(supplier);
    }

    @Override
    default SortedSetX<T> zip(BinaryOperator<Zippable<T>> combiner, final Zippable<T> app) {
        return (SortedSetX<T>)LazyCollectionX.super.zip(combiner,app);
    }

    @Override
    default <R> SortedSetX<R> zipWith(Iterable<Function<? super T, ? extends R>> fn) {
        return (SortedSetX<R>)LazyCollectionX.super.zipWith(fn);
    }

    @Override
    default <R> SortedSetX<R> zipWithS(Stream<Function<? super T, ? extends R>> fn) {
        return (SortedSetX<R>)LazyCollectionX.super.zipWithS(fn);
    }

    @Override
    default <R> SortedSetX<R> zipWithP(Publisher<Function<? super T, ? extends R>> fn) {
        return (SortedSetX<R>)LazyCollectionX.super.zipWithP(fn);
    }

    @Override
    default <T2, R> SortedSetX<R> zipP(final Publisher<? extends T2> publisher, final BiFunction<? super T, ? super T2, ? extends R> fn) {
        return (SortedSetX<R>)LazyCollectionX.super.zipP(publisher,fn);
    }



    @Override
    default <U> SortedSetX<Tuple2<T, U>> zipP(final Publisher<? extends U> other) {
        return (SortedSetX)LazyCollectionX.super.zipP(other);
    }


    @Override
    default <S, U, R> SortedSetX<R> zip3(final Iterable<? extends S> second, final Iterable<? extends U> third, final Fn3<? super T, ? super S, ? super U, ? extends R> fn3) {
        return (SortedSetX<R>)LazyCollectionX.super.zip3(second,third,fn3);
    }

    @Override
    default <T2, T3, T4, R> SortedSetX<R> zip4(final Iterable<? extends T2> second, final Iterable<? extends T3> third, final Iterable<? extends T4> fourth, final Fn4<? super T, ? super T2, ? super T3, ? super T4, ? extends R> fn) {
        return (SortedSetX<R>)LazyCollectionX.super.zip4(second,third,fourth,fn);
    }

    /**
     * Narrow a covariant SortedSet
     * 
     * <pre>
     * {@code 
     * SortedSetX<? extends Fruit> set = SortedSetX.of(apple,bannana);
     * SortedSetX<Fruit> fruitSet = SortedSetX.narrowK3(set);
     * }
     * </pre>
     * 
     * @param sortedSetX to narrowK3 generic type
     * @return SortedSetX with narrowed type
     */
    public static <T> SortedSetX<T> narrow(final SortedSetX<? extends T> setX) {
        return (SortedSetX<T>) setX;
    }

    static class Comparables {

        static <T, R extends ReactiveSeq<T> & Comparable<T>> R comparable(final Seq<T> seq) {
            return comparable(ReactiveSeq.fromStream(seq));
        }

        @SuppressWarnings("unchecked")

        static <T, R extends ReactiveSeq<T> & Comparable<T>> R comparable(final ReactiveSeq<T> seq) {
            final Method compareTo = Stream.of(Comparable.class.getMethods())
                                           .filter(m -> m.getName()
                                                         .equals("compareTo"))
                                           .findFirst()
                                           .get();

            return (R) Proxy.newProxyInstance(SortedSetX.class.getClassLoader(), new Class[] { ReactiveSeq.class, Comparable.class },
                                              (proxy, method, args) -> {
                                                  if (compareTo.equals(method))
                                                      return Objects.compare(System.identityHashCode(seq), System.identityHashCode(args[0]),
                                                                             Comparator.naturalOrder());
                                                  else
                                                      return method.invoke(seq, args);
                                              });

        }
    }

    public static  <T,R> SortedSetX<R> tailRec(T initial, Function<? super T, ? extends SortedSetX<? extends Xor<T, R>>> fn) {
        ListX<Xor<T, R>> lazy = ListX.of(Xor.secondary(initial));
        ListX<Xor<T, R>> next = lazy.eager();
        boolean newValue[] = {true};
        for(;;){

            next = next.flatMap(e -> e.visit(s -> {
                        newValue[0]=true;
                        return fn.apply(s); },
                    p -> {
                        newValue[0]=false;
                        return ListX.of(e);
                    }));
            if(!newValue[0])
                break;

        }
        return Xor.sequencePrimary(next).get().to().sortedSetX(Evaluation.LAZY);
    }

}
